local PassLoot = LibStub("AceAddon-3.0"):GetAddon("PassLoot")
local module = PassLoot:NewModule("Pawn", "AceEvent-3.0")
if not module then return end

local PL = LibStub("AceLocale-3.0"):GetLocale("PassLoot_Modules")
local L = LibStub("AceLocale-3.0"):GetLocale("PassLoot_Pawn")

local function HasPawn()
  -- return PawnCommon and PawnCommon.Scales and PawnCommon.Digits and PawnIsScaleVisible and PawnGetItemData and PawnGetItemDataForInventorySlot
  return PawnVersion and PawnVersion > 2
end

module.Choices = {
  {
    ["Name"] = PL["Any"],
    ["Text"] = PL["Any"],
    ["Value"] = 1,
  },
  {
    ["Name"] = PL["Equal to"],
    ["Text"] = PL["Equal to %num%"],
    ["Value"] = 2,
  },
  {
    ["Name"] = PL["Not Equal to"],
    ["Text"] = PL["Not Equal to %num%"],
    ["Value"] = 3,
  },
  {
    ["Name"] = PL["Less than"],
    ["Text"] = PL["Less than %num%"],
    ["Value"] = 4,
  },
  {
    ["Name"] = PL["Greater than"],
    ["Text"] = PL["Greater than %num%"],
    ["Value"] = 5,
  },
}
module.EquipSlotToInvNumber = {
  [""] = nil,
  ["INVTYPE_AMMO"] = { 0 },
  ["INVTYPE_HEAD"] = { 1 },
  ["INVTYPE_NECK"] = { 2 },
  ["INVTYPE_SHOULDER"] = { 3 },
  ["INVTYPE_BODY"] = { 4 },
  ["INVTYPE_CHEST"] = { 5 },
  ["INVTYPE_ROBE"] = { 5 },
  ["INVTYPE_WAIST"] = { 6 },
  ["INVTYPE_LEGS"] = { 7 },
  ["INVTYPE_FEET"] = { 8 },
  ["INVTYPE_WRIST"] = { 9 },
  ["INVTYPE_HAND"] = { 10 },
  ["INVTYPE_FINGER"] = { 11, 12 },
  ["INVTYPE_TRINKET"] = { 13, 14 },
  ["INVTYPE_CLOAK"] = { 15 },
  ["INVTYPE_WEAPON"] = { 16, 17 },
  ["INVTYPE_SHIELD"] = { 17 },
  ["INVTYPE_2HWEAPON"] = { 16 },
  ["INVTYPE_WEAPONMAINHAND"] = { 16 },
  ["INVTYPE_WEAPONOFFHAND"] = { 17 },
  ["INVTYPE_HOLDABLE"] = { 17 },
  ["INVTYPE_RANGED"] = { 18 },
  ["INVTYPE_THROWN"] = { 18 },
  ["INVTYPE_RANGEDRIGHT"] = { 18 },
  ["INVTYPE_RELIC"] = { 18 },
  ["INVTYPE_TABARD"] = { 19 },
  ["INVTYPE_BAG"] = { 20, 21, 22, 23 },
  ["INVTYPE_QUIVER"] = { 20, 21, 22, 23 },
}
module.ConfigOptions_RuleDefaults = {
  -- { VariableName, Default },
  { "Pawn",
    -- {
      -- [1] = { Scale, Operator, Comparison, Exception }
    -- },
  },
}
module.NewFilterValue_Scale = -1  -- -1 == Match All, or a string for the scale
module.NewFilterValue_LogicalOperator = 1
module.NewFilterValue_Comparison = 0

function module:OnEnable()
  self:RegisterDefaultVariables(self.ConfigOptions_RuleDefaults)
  self:AddWidget(self.Widget)
  self:CheckDBVersion(2, "UpgradeDatabase")
  self:RegisterMessage("PassLoot_OnProfileChanged")
end

function module:OnDisable()
  self:UnregisterDefaultVariables()
  self:RemoveWidgets()
  self:UnregisterMessage("PassLoot_OnProfileChanged")
end

function module:PassLoot_OnProfileChanged()
  if ( not HasPawn() ) then
    for Key, Value in pairs(PassLoot.db.profile.Rules) do
      local Data = self.Widget:GetData(Key)
      if ( #Data > 0 ) then
        if not StaticPopupDialogs.PASSLOOT_PAWN_NOT_LOADED then
          StaticPopupDialogs.PASSLOOT_PAWN_NOT_LOADED = {
            text = L["Pawn is not loaded, rules with Pawn filters will not be matched."],
            button1 = OKAY,
            timeout = 60,
            whileDead = 1,
            hideOnEscape = 1,
            preferredIndex = 3,
          }
        end
        StaticPopup_Show("PASSLOOT_PAWN_NOT_LOADED")
        return
      end
    end
  end
end

function module:UpgradeDatabase(FromVersion, Rule)
  if ( FromVersion == 1 ) then
    local Table = {
      { "Pawn", nil },
    }
    if ( type(Rule.Pawn) == "table" ) then
      if ( #Rule.Pawn == 0 ) then
        return Table
      end
    end
  end
  return
end

function module:CreateWidget()
  local Frame = CreateFrame("Frame")
  Frame:SetWidth(300)
  Frame:SetHeight(43)

  Frame.Title = Frame:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
  Frame.Title:SetParent(Frame)
  Frame.Title:SetPoint("TOP", Frame, "TOP")
  Frame.Title:SetText(L["Pawn not loaded!"])
  Frame.Title:Hide()

  Frame.DropDown = CreateFrame("Frame", "PassLoot_Frames_Widgets_Pawn", Frame, "UIDropDownMenuTemplate")
  Frame.DropDown:EnableMouse(true)
  Frame.DropDown:SetHitRectInsets(15, 15, 0 ,0)
  _G[Frame.DropDown:GetName().."Text"]:SetJustifyH("CENTER")
  UIDropDownMenu_SetWidth(Frame.DropDown, 280)
  Frame.DropDown:SetScript("OnEnter", function() self:ShowTooltip(L["Pawn Scale"], L["TooltipDesc"]) end)
  Frame.DropDown:SetScript("OnLeave", function() GameTooltip:Hide() end)
  local Button = _G[Frame.DropDown:GetName().."Button"]
  Button:SetScript("OnEnter", function() self:ShowTooltip(L["Pawn Scale"], L["TooltipDesc"]) end)
  Button:SetScript("OnLeave", function() GameTooltip:Hide() end)
  Frame.DropDown.Title = Frame.DropDown:CreateFontString(Frame.DropDown:GetName().."Title", "BACKGROUND", "GameFontNormalSmall")
  Frame.DropDown.Title:SetParent(Frame.DropDown)
  Frame.DropDown.Title:SetPoint("BOTTOMLEFT", Frame.DropDown, "TOPLEFT", 20, 0)
  Frame.DropDown.Title:SetText(L["Pawn Scale"])
  Frame.DropDown.initialize = function(...) self:DropDown_Init(...) end
  Frame.DropDown:SetPoint("TOP", Frame, "TOP", 0, -1 * Frame.DropDown.Title:GetHeight())
  Frame.DropDown:Hide()

  Frame.Height = Frame:GetHeight()
  Frame.Width = Frame:GetWidth()
  Frame.Info = {
    L["FilterListName"],
    L["FilterListDesc"],
  }

  local DropDownEditBox = CreateFrame("EditBox", "PassLoot_Frames_Widgets_PawnDropDownEditBox")
  DropDownEditBox:Hide()
  DropDownEditBox:SetParent(nil)
  DropDownEditBox:SetFontObject(ChatFontNormal)
  DropDownEditBox:SetMaxLetters(50)  -- Was 8
  DropDownEditBox:SetAutoFocus(true)
  DropDownEditBox:SetScript("OnEnter", function(frame)
    CloseDropDownMenus(frame:GetParent():GetParent():GetID() + 1)
    UIDropDownMenu_StopCounting(frame:GetParent():GetParent())
  end)
  DropDownEditBox:SetScript("OnEnterPressed", function(frame)
    self:DropDown_OnClick(frame)  -- Calls Hide(), ClearAllPoints() and SetParent(nil)
    -- CloseMenus() only hides the DropDownList2, not this object, and even tho i will set parent to nil, i might as well cover bases
    CloseMenus()
  end)
  DropDownEditBox:SetScript("OnEscapePressed", function(frame)
    frame:Hide()
    frame:ClearAllPoints()
    frame:SetParent(nil)
    CloseMenus()
  end)
  DropDownEditBox:SetScript("OnEditFocusGained", function(frame) UIDropDownMenu_StopCounting(frame:GetParent():GetParent()) end)
  DropDownEditBox:SetScript("OnEditFocusLost", function(frame) frame:Hide() end)
  return Frame, DropDownEditBox
end
module.Widget, module.DropDownEditBox = module:CreateWidget()

-- Local function to get the data and make sure it's valid data
function module.Widget:GetData(RuleNum)
  local Data = module:GetConfigOption("Pawn", RuleNum)
  local Changed = false
  if ( Data ) then
    if ( type(Data) == "table" and #Data > 0 ) then
      for Key, Value in ipairs(Data) do
        if ( type(Value) ~= "table" or (type(Value[1]) ~= "number" and type(Value[1]) ~= "string") or type(Value[2]) ~= "number" or (type(Value[3]) ~= "number" and type(Value[3]) ~= "string") ) then
          Data[Key] = {
            module.NewFilterValue_Scale,
            module.NewFilterValue_LogicalOperator,
            module.NewFilterValue_Comparison,
            false
          }
          Changed = true
        end
      end
    else
      Data = nil
      Changed = true
    end
  end
  if ( Changed ) then
    module:SetConfigOption("Pawn", Data)
  end
  return Data or {}
end

function module.Widget:GetNumFilters(RuleNum)
  local Value = self:GetData(RuleNum)
  return #Value
end

function module.Widget:AddNewFilter()
  local Value = self:GetData()
  local NewTable = {
    module.NewFilterValue_Scale,
    module.NewFilterValue_LogicalOperator,
    module.NewFilterValue_Comparison,
    false
  }
  table.insert(Value, NewTable)
  module:SetConfigOption("Pawn", Value)
end

function module.Widget:RemoveFilter(Index)
  local Value = self:GetData()
  table.remove(Value, Index)
  if ( #Value == 0 ) then
    Value = nil
  end
  module:SetConfigOption("Pawn", Value )
end

function module.Widget:DisplayWidget(Index)
  if ( HasPawn() ) then
    self.Title:Hide()
    self.DropDown:Show()
    if ( Index ) then
      module.FilterIndex = Index
    end
    local Value = self:GetData()
    local Value_Scale = Value[module.FilterIndex][1]
    local Value_LogicalOperator = Value[module.FilterIndex][2]
    local Value_Comparison = Value[module.FilterIndex][3]
    UIDropDownMenu_SetText(module.Widget.DropDown, module:GetPawnText(Value_Scale, Value_LogicalOperator, Value_Comparison))
  else
    self.Title:Show()
    self.DropDown:Hide()
  end
end

function module.Widget:GetFilterText(Index)
  local Value = self:GetData()
  if ( Value[Index][1] and Value[Index][2] and Value[Index][3] ) then
    if ( HasPawn() ) then
      return module:GetPawnText(Value[Index][1], Value[Index][2], Value[Index][3])
    else
      return L["Pawn not loaded!"]
    end
  end
end

function module.Widget:IsException(RuleNum, Index)
  local Data = self:GetData(RuleNum)
  return Data[Index][4]
end

function module.Widget:SetException(RuleNum, Index, Value)
  local Data = self:GetData(RuleNum)
  Data[Index][4] = Value
  module:SetConfigOption("Pawn", Data)
end

module.CurrentMatch = {}
module.CurrentMatch_EquippedItems = {}
function module.Widget:SetMatch(ItemLink, Tooltip)
  for Key, Value in pairs(module.CurrentMatch) do
    module.CurrentMatch[Key] = nil
  end
  for Key, Value in pairs(module.CurrentMatch_EquippedItems) do
    module.CurrentMatch_EquippedItems[Key] = nil
  end
  if ( Scale == -1 ) then  -- No scale selected
    return
  end

  local _, _, _, _, _, _, _, _, EquipSlot, _ = GetItemInfo(ItemLink)
  local InvSlot = module.EquipSlotToInvNumber[EquipSlot]
  if ( InvSlot and HasPawn() ) then
    module.CurrentMatch = PawnGetItemData(ItemLink)
    for Key, Value in pairs(InvSlot) do
      if ( GetInventoryItemLink("player", Value) ) then
        table.insert(module.CurrentMatch_EquippedItems, PawnGetItemDataForInventorySlot(Value, false, "player"))
      end
    end
  end
end

function module.Widget:GetMatch(RuleNum, Index)
  local RuleValue = self:GetData(RuleNum)
  local Scale = RuleValue[Index][1]
  local LogicalOperator = RuleValue[Index][2]
  if ( not HasPawn() ) then
    return false
  end
  if ( Scale == -1 ) then  -- No scale selected.
    return true
  end

  local CurrentMatch
  -- First get the Pawn value of the item that dropped for the scale that was selected.
  local PawnData = module.CurrentMatch
  if ( PawnData.Values ) then
    for ListKey, ListValue in pairs(PawnData.Values) do
      local ScaleName, Value, UnenchantedValue, LocalizedName = ListValue[1], ListValue[2], ListValue[3], ListValue[7]
      if ( PawnCommon.Scales[ScaleName] and ScaleName == Scale ) then
        CurrentMatch = UnenchantedValue
        break
      end
    end
  else
    module:Debug("No pawn data")
    return false
  end
  CurrentMatch = CurrentMatch or 0  -- Pawn doesn't add a scale to the list if it's value is 0
  -- Now get the pawn values of the items that are taking up the same inventory slot(s) for the scale that was selected.
  local TempNum
  for EquippedKey, EquippedValue in pairs(module.CurrentMatch_EquippedItems) do
    if ( EquippedValue.Values ) then
      for ListKey, ListValue in pairs(EquippedValue.Values) do
        local ScaleName, Value, UnenchantedValue, LocalizedName = ListValue[1], ListValue[2], ListValue[3], ListValue[7]
        if ( PawnCommon.Scales[ScaleName] and ScaleName == Scale ) then
          if ( not TempNum ) then
            TempNum = UnenchantedValue
          else
            TempNum = math.min(UnenchantedValue, TempNum)
          end
          break
        end
      end
    end
  end
  TempNum = TempNum or 0
  local Digits = 10 ^ PawnCommon.Digits
  CurrentMatch = math.floor(CurrentMatch * Digits + 0.5) / Digits
  TempNum = math.floor(TempNum * Digits + 0.5) / Digits
  module:Debug("Drop: "..CurrentMatch.." Equipped: "..TempNum)
  -- Evaluate the Comparison using TempNum as 'current' keyword.
  self.Environment[L["current"]] = TempNum
  local Comparison = self:Evaluate(RuleValue[Index][3])
  if ( not Comparison ) then
    return false
  end

  if ( LogicalOperator > 1 ) then
    if ( LogicalOperator == 2 ) then -- Equal To
      if ( CurrentMatch ~= Comparison ) then
        module:Debug(CurrentMatch.." ~= "..Comparison)
        return false
      end
    elseif ( LogicalOperator == 3 ) then -- Not Equal To
      if ( CurrentMatch == Comparison ) then
        module:Debug(CurrentMatch.." == "..Comparison)
        return false
      end
    elseif ( LogicalOperator == 4 ) then -- Less than
      if ( CurrentMatch >= Comparison ) then
        module:Debug(CurrentMatch.." >= "..Comparison)
        return false
      end
    elseif ( LogicalOperator == 5 ) then -- Greater than
      if ( CurrentMatch <= Comparison ) then
        module:Debug(CurrentMatch.." <= "..Comparison)
        return false
      end
    end
  end
  return true
end

-- Create an environment for the functions, so they can't access globals and such.
-- We can also add our variables here, so we don't have to string.gsub
module.Widget.Environment = {
  [L["current"]] = 0,
}
module.Widget.CachedFunc, module.Widget.CachedError = {}, {} -- A list of functions and errors we have already tried loading
function module.Widget:Evaluate(Logic)
  local Function, Error = self.CachedFunc[Logic], self.CachedError[Logic]
  if ( not Function and not Error ) then
    Function, Error = loadstring("return "..Logic)
    self.CachedFunc[Logic], self.CachedError[Logic] = Function, Error
  end
  if ( Function and not Error ) then
    setfenv(Function, self.Environment)  -- Limit what the loaded logic string can access
    local Success, ReturnValue = pcall(Function)  -- Catch errors.
    if ( Success ) then
      return ReturnValue
    else
      self:Debug("Could not evaluate "..(Logic or "").." - "..(ReturnValue or ""))
      return
    end
  else
    self:Debug("Could not evaluate "..(Logic or "").." - "..(Error or ""))
  end
end

local SortedList = {}
function module:DropDown_Init(Frame, Level)
  Level = Level or 1
  local info = {}
  info.checked = false
  info.notCheckable = true
  info.func = function(...) self:DropDown_OnClick(...) end
  info.owner = Frame
  local List = PawnCommon.Scales
  if ( Level == 1 ) then
    info.text = L["Match All"]
    info.hasArrow = false
    info.notClickable = false
    info.value = {
      ["Scale"] = -1,
      ["Operator"] = 1,
    }
    UIDropDownMenu_AddButton(info, Level)
    for Key, Value in pairs(SortedList) do
      SortedList[Key] = nil
    end
    for Key, Value in pairs(List) do
      if ( PawnCommon.Scales[Key] and PawnIsScaleVisible(Key) ) then
        table.insert(SortedList, Key)
      end
    end
    table.sort(SortedList)
    for Key, Value in ipairs(SortedList) do
      info.text = Value
      info.hasArrow = true
      info.notClickable = false
      info.value = {
        ["Scale"] = Value,
        ["Operator"] = 1,
      }
      UIDropDownMenu_AddButton(info, Level)
    end
  elseif ( Level == 2 ) then
    for Key, Value in ipairs(self.Choices) do
      info.text = Value.Name
      info.value = {
        ["Scale"] = UIDROPDOWNMENU_MENU_VALUE.Scale,
        ["Operator"] = Value.Value,
      }
      if ( Key == 1 ) then
        info.hasArrow = false
      else
        info.hasArrow = true
      end
      info.notClickable = false
      UIDropDownMenu_AddButton(info, Level)
    end
  else
    self.DropDownEditBox.value = UIDROPDOWNMENU_MENU_VALUE
    info.text = ""
    info.notClickable = false
    UIDropDownMenu_AddButton(info, Level)
    DropDownList3.maxWidth = 80
    DropDownList3:SetWidth(80)
    self.DropDownEditBox.owner = info.owner
    self.DropDownEditBox:ClearAllPoints()
    self.DropDownEditBox:SetParent(DropDownList3Button1)
    self.DropDownEditBox:SetPoint("TOPLEFT", DropDownList3Button1, "TOPLEFT")
    self.DropDownEditBox:SetPoint("BOTTOMRIGHT", DropDownList3Button1, "BOTTOMRIGHT")
    local Value = self.Widget:GetData()
    self.DropDownEditBox:SetText(Value[self.FilterIndex][3])
    self.DropDownEditBox:Show()
    self.DropDownEditBox:SetFocus()
    self.DropDownEditBox:HighlightText()
  end
end

function module:DropDown_OnClick(Frame)
  local Value = self.Widget:GetData()
  local Scale = Frame.value.Scale
  local LogicalOperator = Frame.value.Operator
  local Comparison = Value[self.FilterIndex][3]
  if ( Frame:GetName() == self.DropDownEditBox:GetName() ) then
    Comparison = Frame:GetText() or ""
  end
  Value[self.FilterIndex][1] = Scale
  Value[self.FilterIndex][2] = LogicalOperator
  Value[self.FilterIndex][3] = Comparison
  self:SetConfigOption("Pawn", Value)
  UIDropDownMenu_SetText(Frame.owner, self:GetPawnText(Scale, LogicalOperator, Comparison))
  self.DropDownEditBox:Hide()
  self.DropDownEditBox:ClearAllPoints()
  self.DropDownEditBox:SetParent(nil)
end

function module:GetPawnText(Scale, LogicalOperator, Comparison)
  local ComparisonText
  if ( Scale == -1 ) then
    return L["Match All"]
  end
  for Key, Value in ipairs(self.Choices) do
    if ( Value.Value == LogicalOperator ) then
      ComparisonText = string.gsub(Value.Text, "%%num%%", Comparison)
      break
    end
  end
  local Text = string.gsub(L["FILTER_TEXT"], "%%scale%%", Scale)
  Text = string.gsub(Text, "%%comparison%%", ComparisonText)
  return Text
end
